searchData={"items":[{"type":"module","title":"Efx","doc":"Testing with side-effects is often hard. Various solutions exist to work around\nthe difficulties, e.g. mocking. This library offers a very easy way to achieve \ntestable code by mocking. Instead of mocking we talk about binding effects to another implementation.\n`Efx` offers a declarative way to mark effectful functions and bind them in tests.","ref":"Efx.html"},{"type":"module","title":"Rationale - Efx","doc":"Efx is a small library that does one thing and one thing only very well: Make code\nthat contains side effects testable. \n\nExisting mock libraries often set up mocks in non declarative ways: configs need \nto be adapted & mock need to be initialized. In source code there are intrusive \ninstructions to set up mockable code. `Efx` is very unintrusive in both, source\ncode and test code. It offers a convenient and declarative syntax. Instead of \nmocking we talk about binding effects.\n\nEfx follows the following principles:\n\n- Implementing and binding effects should be as simple and declarative as possible.\n- Modules contain groups of effects that can only be bound as a set.\n- We want to run as much tests async as possible. Thus, we traverse \n  the supervision tree to find rebound effects in the ancest test processes,\n  in an isolated manner.\n- Effects by default execute their default implemenation in tests, and thus, must be explicitly bound.\n- Effects can only be bound in tests, but not in production. In production always the default implementation is executed.\n- We want zero performance overhead in production.","ref":"Efx.html#module-rationale"},{"type":"module","title":"Usage - Efx","doc":"","ref":"Efx.html#module-usage"},{"type":"module","title":"Defining Effects - Efx","doc":"An effect is a function that executes side effects and thus is\nhard to test or even untestable. With the effects abstraction we\ncan define mockable effect-functions comfortably.\n\nSince effects are used on a module level utilizing the use-macro,\nall effect-functions defined inside build a group. We can later\neither bind all or none of this functions.\n\nThis module provides macros for implementing effects that are expanded into a\nmockable behaviour, e.g.\n\n\n```elixir\ndefmodule MyEffect do\n  use Efx\n\n  @spec read_numbers(String.t()) :: integer()\n  defeffect read_numbers(id) do\n    ... \n  end\n\n  @spec write_numbers(String.t(), integer()) :: :ok\n  defeffect read_numbers(id, numbers) do\n    ...\n  end\nend\n```\n\n\n  The above example generates a behaviour with the callbacks\n\n```elixir\n      @callback read_numbers(String.t()) :: integer()\n      @callback write_numbers(String.t(), integer()) :: :ok\n```\n\nBy using the `defeffect`-macro, we define an effect-function as well as provide \na default-implementation in its body. For more detail see the moduledoc in the\n`Efx`-module.","ref":"Efx.html#module-defining-effects"},{"type":"module","title":"Binding Effects in Tests - Efx","doc":"To bind effects one simply has to use `EfxCase`-Module and call bind functions. Lets say we have the following effects implementation:\n\n```elixir\ndefmodule MyModule do\n  use Common.Effects \n\n  @spec get() :: list()\n  defeffect get() do\n     ...\n  end\nend\n```\n  \nThe following shows code that binds the effect to a different implementation in tests:\n\n```elixir\ndefmodule SomeTest do\n  use Common.EffectsCase\n\n  test \"test something\" do\n    bind(MyModule, :get, fn -> [1,2,3] end)\n    ...\n  end\nend\n```\n\nInstead of returning the value of the default implementation, `MyModule.get/0` returns `[1,2,3]`.\n\nFor more details see the `EfxCase`-module.","ref":"Efx.html#module-binding-effects-in-tests"},{"type":"macro","title":"Efx.defeffect/2","doc":"","ref":"Efx.html#defeffect/2"},{"type":"module","title":"Efx.Application","doc":"","ref":"Efx.Application.html"},{"type":"function","title":"Efx.Application.start/2","doc":"","ref":"Efx.Application.html#start/2"},{"type":"module","title":"EfxCase","doc":"Module for testing with effects.\n\nBinding effects in tests follows these principles:\n\n- By default, all effects of a module is bound to the default implementation.\n- We either bind all effect functions of a module or none.\n  We cannot bind single functions (except the explicit use of :default).\n  If we rebind only one effect and the other is called, we raise. \n- A function is either bound without or with a specified number of expected calls.\n  If a function has multiple binds, they are called in given order, until they satisfied their\n  expected number of calls.\n- The number of expected calls is always veryified.","ref":"EfxCase.html"},{"type":"module","title":"Binding effects - EfxCase","doc":"To bind effects one simply has to use this module and call\nthe bind macro. Lets say we have the following effects implementation:\n\n    defmodule MyModule do\n      use Common.Effects \n  \n      @spec get() :: list()\n      defeffect get() do\n         ...\n      end\n    end\n  \nThe following shows code that binds the effect:\n\n    defmodule SomeTest do\n      use Common.EffectsCase\n  \n      test \"test something\" do\n        bind(MyModule, :get, fn -> [1,2,3] end)\n        ...\n      end\n    end\n\nInstead of returning the value of the default implementation,\n`MyModule.get/0` returns `[1,2,3]`.","ref":"EfxCase.html#module-binding-effects"},{"type":"module","title":"Binding with an expected Number of Calls - EfxCase","doc":"We can additionally define an expected number of calls. The expected\nnumber of calls is always verified - a test run will fail if\nit is not satisfied, as well as exceeded.\n\nWe can define a number of expected calls as follows:\n\n\n    defmodule SomeTest do\n      use Common.EffectsCase\n  \n      test \"test something\" do\n        expect(MyModule, :get, fn -> [1,2,3] end, calls: 2)\n        ...\n      end\n    end\n\nIn this case, we verify that the bound function `get/0` is called\nexactly twice.","ref":"EfxCase.html#module-binding-with-an-expected-number-of-calls"},{"type":"module","title":"Binding globally - EfxCase","doc":"Effect binding uses process dictionaries to find the right binding\nthrough-out the supervision-tree.\nAs long as calling processes have the testing process that defines\nthe binding as an ancestor, binding works. If we cannot ensure that,\nwe can set binding to global. However, then the tests must be set\nto async to not interfere:\n\n    defmodule SomeTest do\n      use Common.EffectsCase, async: false\n  \n      test \"test something\" do\n        bind(MyModule, :get, fn -> [1,2,3] end)\n        ...\n      end\n    end","ref":"EfxCase.html#module-binding-globally"},{"type":"module","title":"Binding the same Function with multiple bind-Calls - EfxCase","doc":"We can chain binds for the same functions. They then\nget executed until their number of expected calls is satisfied:\n\n    defmodule SomeTest do\n      use Common.EffectsCase\n  \n      test \"test something\" do\n        bind(MyModule, :get,  fn -> [1,2,3] end, calls: 1)\n        bind(MyModule, :get,  fn -> [] end, calls: 2)\n        bind(MyModule, :get, fn -> [1,2] end)\n        ...\n      end\n    end\n\nIn this example the first binding of `get/0` gets called one time,\nthen the second binding is used to replace the call two more times\nand the last get, specified without an expected number of calls,\nis used for the rest of the execution.","ref":"EfxCase.html#module-binding-the-same-function-with-multiple-bind-calls"},{"type":"module","title":"Setup for many Tests - EfxCase","doc":"If we want to setup the same binding for multiple tests we can do\nthis as follows:\n\n    defmodule SomeTest do\n      use Common.EffectsCase\n\n      setup_effects(MyModule,\n         :get, fn -> [1,2,3] end\n      )\n  \n      test \"test something\" do\n        # test with mocked get\n        ...\n      end\n    end","ref":"EfxCase.html#module-setup-for-many-tests"},{"type":"module","title":"Explicitly defaulting one Function in Tests - EfxCase","doc":"While it is best practice to bind all function of a module or none,\nwe can also default certain functions explicitly:\n\n    defmodule MyModule do\n      use Common.Effects \n  \n      @spec get() :: list()\n      defeffect get() do\n         ...\n      end\n  \n      @spec put(any()) :: :ok\n      defeffect put(any()) do\n         ...\n      end\n    end\n  \n\n    defmodule SomeTest do\n      use Common.EffectsCase\n  \n      test \"test something\" do\n        bind(MyModule, :get, fn -> [1,2,3] end)\n        bind(MyModule, :put, {:default, 0})\n        ...\n      end\n    end\n\nWhile entirely leaving out `put/1` would result in an error\n(when called), we can tell the effects library to use it's\ndefault implementation. Note that defaulting can be combined\nwith an expected number of calls.","ref":"EfxCase.html#module-explicitly-defaulting-one-function-in-tests"},{"type":"function","title":"EfxCase.bind/5","doc":"","ref":"EfxCase.html#bind/5"},{"type":"macro","title":"EfxCase.setup_effects/2","doc":"","ref":"EfxCase.html#setup_effects/2"},{"type":"module","title":"EfxCase.Internal","doc":"","ref":"EfxCase.Internal.html"},{"type":"function","title":"EfxCase.Internal.init/1","doc":"","ref":"EfxCase.Internal.html#init/1"},{"type":"function","title":"EfxCase.Internal.verify_mocks!/1","doc":"","ref":"EfxCase.Internal.html#verify_mocks!/1"},{"type":"module","title":"EfxCase.Mock","doc":"Internal logic of a mocked behaviour. This module is meant\nto make handling a mock, counting the calls, finding the\nright mocked function, calling the mocked function, etc... more\nconvenient.\n\nA mock consists of the following:\n\n- a list of mocked functions","ref":"EfxCase.Mock.html"},{"type":"function","title":"EfxCase.Mock.add_fun/5","doc":"","ref":"EfxCase.Mock.html#add_fun/5"},{"type":"function","title":"EfxCase.Mock.get_fun/3","doc":"","ref":"EfxCase.Mock.html#get_fun/3"},{"type":"function","title":"EfxCase.Mock.get_unsatisfied/1","doc":"","ref":"EfxCase.Mock.html#get_unsatisfied/1"},{"type":"function","title":"EfxCase.Mock.inc_called/3","doc":"","ref":"EfxCase.Mock.html#inc_called/3"},{"type":"function","title":"EfxCase.Mock.make/1","doc":"","ref":"EfxCase.Mock.html#make/1"},{"type":"type","title":"EfxCase.Mock.t/0","doc":"","ref":"EfxCase.Mock.html#t:t/0"},{"type":"module","title":"EfxCase.Mock.MockedFun","doc":"Internal logic of a mocked function. \n\nA mocked functions consists of the following:\n\n- the name or identifier of the function\n- the arity of the function\n- the implementation/replacement of the function. Besides an\n  anonymous function this can be :unmocked or :default.\n  :unmocked says there is no replacement yet and error when called \n  :default refers to the default implementation\n- a number of expected calls\n- a counter showing how the mocked functions has been called","ref":"EfxCase.Mock.MockedFun.html"},{"type":"function","title":"EfxCase.Mock.MockedFun.limit_reached?/1","doc":"","ref":"EfxCase.Mock.MockedFun.html#limit_reached?/1"},{"type":"function","title":"EfxCase.Mock.MockedFun.satisfied?/1","doc":"","ref":"EfxCase.Mock.MockedFun.html#satisfied?/1"},{"type":"type","title":"EfxCase.Mock.MockedFun.t/0","doc":"","ref":"EfxCase.Mock.MockedFun.html#t:t/0"},{"type":"module","title":"EfxCase.MockState","doc":"This module implements a global mutable state needed for mocking.\n\nMocking works by adding the pid of the mock-owning process to it's\ndictionary and accessing this pid using a lookup with `ProcessTree`.\n\nThe pid then can be used to lookup information of the mock, e.g. which\nfunction is mocked with replacement or how often a function was\ncalled to verify. This module is the global lookup.\n\nWe need a global lookup in which we can manipulate data to count mock\ninvocations as well as store mocks as global when a test is flagged as\nasync.","ref":"EfxCase.MockState.html"},{"type":"function","title":"EfxCase.MockState.add_fun/5","doc":"","ref":"EfxCase.MockState.html#add_fun/5"},{"type":"function","title":"EfxCase.MockState.add_fun/6","doc":"","ref":"EfxCase.MockState.html#add_fun/6"},{"type":"function","title":"EfxCase.MockState.call/3","doc":"","ref":"EfxCase.MockState.html#call/3"},{"type":"function","title":"EfxCase.MockState.call/4","doc":"","ref":"EfxCase.MockState.html#call/4"},{"type":"function","title":"EfxCase.MockState.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EfxCase.MockState.html#child_spec/1"},{"type":"function","title":"EfxCase.MockState.clean_globals/0","doc":"","ref":"EfxCase.MockState.html#clean_globals/0"},{"type":"function","title":"EfxCase.MockState.mocked?/1","doc":"","ref":"EfxCase.MockState.html#mocked?/1"},{"type":"function","title":"EfxCase.MockState.mocked?/2","doc":"","ref":"EfxCase.MockState.html#mocked?/2"},{"type":"function","title":"EfxCase.MockState.parse/1","doc":"","ref":"EfxCase.MockState.html#parse/1"},{"type":"function","title":"EfxCase.MockState.start_link/1","doc":"","ref":"EfxCase.MockState.html#start_link/1"},{"type":"function","title":"EfxCase.MockState.verify_called!/0","doc":"","ref":"EfxCase.MockState.html#verify_called!/0"},{"type":"function","title":"EfxCase.MockState.verify_called!/1","doc":"","ref":"EfxCase.MockState.html#verify_called!/1"},{"type":"type","title":"EfxCase.MockState.pid_t/0","doc":"","ref":"EfxCase.MockState.html#t:pid_t/0"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}